@@ -1,4 +1,9 @@
 """Shared models."""
+from obsidian_metadata.models.enums import (
+    InsertLocation,
+    MetadataType,
+)
+
 from obsidian_metadata.models.patterns import Patterns  # isort: skip
 from obsidian_metadata.models.metadata import (
     Frontmatter,
@@ -7,18 +12,21 @@ from obsidian_metadata.models.metadata import (
     VaultMetadata,
 )
 from obsidian_metadata.models.notes import Note
-from obsidian_metadata.models.vault import Vault
+from obsidian_metadata.models.vault import Vault, VaultFilter
 
 from obsidian_metadata.models.application import Application  # isort: skip
 
 __all__ = [
+    "Application",
     "Frontmatter",
     "InlineMetadata",
     "InlineTags",
+    "InsertLocation",
     "LoggerManager",
+    "MetadataType",
     "Note",
     "Patterns",
-    "Application",
     "Vault",
+    "VaultFilter",
     "VaultMetadata",
 ]
